(* Interfaces and implementations of dictionaries.  A dictionary
 * is used to associate a value with a key.  In our case, we will
 * be using a dictionary to build an index for the web, associating
 * a set of URLs with each word that we find as we crawl the web.
 *)
exception TODO
exception BADSTATE

module type DICT = 
sig
  type key
  type value
  type dict

  (* An empty dictionary *)
  val empty : dict 

  (* Reduce the dictionary using the provided function f and base case u. 
   * Our reducing function f must have the type:
   *      key -> value -> 'a -> 'a
   * and our base case u has type 'a.
   * 
   * If our dictionary is the (key,value) pairs (in any order)
   *      (k1,v1), (k2,v2), (k3,v3), ... (kn,vn)
   * then fold should return:
   *      f k1 v1 (f k2 v2 (f k3 v3 (f ... (f kn vn u))))
  *)

  type rep =
    | RLeaf
    | RNode of rep * key * value * rep

  val fold : (key -> value -> 'a -> 'a) -> 'a -> dict -> 'a
 
  (* Returns as an option the value associated with the provided key. If
   * the key is not in the dictionary, return None. *)
  val lookup : dict -> key -> value option

  (* Returns true if and only if the key is in the dictionary. *)
  val member : dict -> key -> bool

  (* Inserts a (key,value) pair into our dictionary. If the key is already
   * in our dictionary, update the key to have the new value. *)
  val insert : dict -> key -> value -> dict

  (* Removes the given key from the dictionary. If the key is not present,
   * return the original dictionary. *)
  val remove : dict -> key -> dict

  (* Return an arbitrary key, value pair along with a new dict with that
   * pair removed. Return None if the input dict is empty *)
  val choose : dict -> (key * value * dict) option

  (* Return a generator for arbitrary dictionarys *)
  val arbitrary : unit -> (dict QCheck.Gen.t)

  (* Return a generator for arbitrary keys *)
  val arbitrary_key : unit -> (key QCheck.Gen.t)

  (* Return a generator for arbitrary values *)
  val arbitrary_value : unit -> (value QCheck.Gen.t)

  (* When given a dictionary, return a list of smaller sub dictionaries*)
  val shrink : dict -> (dict list)

  (* functions to convert our types to strings for debugging and logging *)
  val string_of_key: key -> string
  val string_of_value : value -> string
  val string_of_dict : dict -> string

  (* functions to expose the internal structure of our dict *)
  val expose:  dict -> rep
  val hide: rep -> dict

  (* Runs all the tests. see "Testing" in the assignment web page *)
  val run_tests : unit -> unit

end



(* Argument module signature to our DICT functors *)
module type DICT_ARG =
sig
  type key
  type value
  val compare : key -> key -> Order.order
  val string_of_key : key -> string
  val string_of_value : value -> string

  (* Use these functions for testing--see "Testing" in the assignment web page *)

  (* Generate a key. The same key is always returned *)
  val gen_key : unit -> key

  (* Generate a random key. *)
  val gen_key_random : unit -> key

  (* Generates a key greater than the argument. *)
  val gen_key_gt : key -> unit -> key

  (* Generates a key less than the argument. *)
  val gen_key_lt : key -> unit -> key

  (* Generates a key between the two arguments. Return None if no such
   * key exists. *)
  val gen_key_between : key -> key -> unit -> key option

  (* Generates a random value. *)
  val gen_value : unit -> value

  (* Generates a random (key,value) pair *)
  val gen_pair : unit -> key * value

  (* Return a generator for arbitrary keys *)
  val arbitrary_key : unit -> (key QCheck.Gen.t)

  (* Return a generator for arbitrary values *)
  val arbitrary_value : unit -> (value QCheck.Gen.t)


end



(* An example implementation of our DICT_ARG signature. Use this struct
 * for testing. *)
module IntStringDictArg : DICT_ARG =
struct
  open Order
  type key = int
  type value = string
  let compare x y = if x < y then Less else if x > y then Greater else Eq
  let string_of_key = string_of_int
  let string_of_value v = v
  let gen_key () = 0
  let gen_key_gt x () = x + 1
  let gen_key_lt x () = x - 1 
  let gen_key_between x y () = 
    let (lower, higher) = (min x y, max x y) in
    if higher - lower < 2 then None else Some (higher - 1)
  let gen_key_random =
    let _ = Random.self_init () in
    (fun () -> Random.int 10000)

  let arbitrary_key () = QCheck.Gen.map (fun (x : key) -> x) QCheck.Gen.(nat)
  let arbitrary_value () = QCheck.Gen.map (fun (x : value) -> x) QCheck.Gen.(small_string)
 
  (* returns the nth string in lst, or "cow" n > length of list *)
  let rec lst_n (lst: string list) (n: int) : string =
    match lst with
      | [] -> "cow"
      | hd::tl -> if n = 0 then hd else lst_n tl (n-1)

  (* list of possible values to generate *)
  let possible_values = ["a";"c";"d";"e";"f";"g";"h";"i";"j";"k";"m";"n";
                         "o";"p";"q";"r";"s";"t";"u";"v";"w";"x";"y";"z";
                         "zzzzzz";"cheese";"foo";"bar";"baz";"quux";"42"]
  let num_values = List.length possible_values
  (* gen_value will return the string at this current index *)
  let current_index = ref 0
  let gen_value () =
    let index = !current_index in
    if index >= num_values then
      (current_index := 0; lst_n possible_values index)
    else
      (current_index := index + 1; lst_n possible_values index)
  let gen_pair () = (gen_key_random(), gen_value())
end



(* An association list implementation of our DICT signature. *)
module AssocListDict(D:DICT_ARG) : (DICT with type key = D.key
  with type value = D.value) = 
struct
  open Order;;
  type key = D.key;;
  type value = D.value;;
  type dict = (key * value) list;;

  type rep =
    | RLeaf
    | RNode of rep * key * value * rep

  (* INVARIANT: sorted by key, no duplicates *)

  let empty = [] ;;

  let fold f u = List.fold_left (fun a (k,v) -> f k v a) u

  let rec lookup d k = 
    match d with 
      | [] -> None
      | (k1,v1)::d1 -> 
        (match D.compare k k1 with
          | Eq -> Some v1
          | Greater -> lookup d1 k 
          | _ -> None)

  let member d k = 
    match lookup d k with 
      | None -> false 
      | Some _ -> true

  let rec insert d k v = 
    match d with 
      | [] -> [(k,v)]
      | (k1,v1)::d1 -> 
        (match D.compare k k1 with 
          | Less -> (k,v)::d
          | Eq -> (k,v)::d1
          | Greater -> (k1,v1)::(insert d1 k v))

  let rec remove d k = 
    match d with 
      | [] -> []
      | (k1,v1)::d1 ->
	(match D.compare k k1 with 
          | Eq -> d1
          | Greater -> (k1,v1)::(remove d1 k)
          | _ -> d)
	  
  let choose d = 
    match d with 
      | [] -> None
      | (k,v)::rest -> Some(k,v,rest)

  let string_of_key = D.string_of_key
  let string_of_value = D.string_of_value
  let string_of_dict (d: dict) : string = 
    let f = (fun y (k,v) -> y ^ "\n key: " ^ D.string_of_key k ^ 
      "; value: (" ^ D.string_of_value v ^ ")") in
    List.fold_left f "" d

  let arbitrary_key () : (key QCheck.Gen.t) = D.arbitrary_key ()

  let arbitrary_value () : (value QCheck.Gen.t) = D.arbitrary_value ()

  let shrink (d:dict) : dict list=
      let rec iter d_sub = match d_sub with
      | a::b -> (remove d (fst a)):: (iter b)
      | [] -> [] in
      (iter d)

  let arbitrary () : (dict QCheck.Gen.t) =
    let gen_pair = QCheck.Gen.map2 (fun x y -> (x,y)) (D.arbitrary_key ()) (D.arbitrary_value ()) in
    let unsorted = QCheck.Gen.list gen_pair in
    let sort_func s1 s2 = match D.compare (fst s1) (fst s2) with | Greater -> 1 | Eq -> 0 | Less -> -1 in
    QCheck.Gen.(unsorted >>= (fun x ->return (List.sort_uniq sort_func x)))

  let rec expose (d:dict) : rep =
    match d with
    | [] -> RLeaf
    | (k, v)::b -> RNode (RLeaf, k, v, expose b)

  let rec hide (r:rep) : dict =
    match r with
    | RLeaf -> []
    | RNode (left, key, value, right) -> (hide left)@((key,value)::(hide right))


  let run_tests () = ()

end


(******************************************************************)
(* AVDict: a functor that implements our DICT signature           *)
(* using red black trees                                          *)
(******************************************************************)
(*
module AVLDict(D:DICT_ARG) : (DICT with type key = D.key
with type value = D.value) =
struct
  open Order

  exception TODO

  (* A dictionary entry is a (key,value) pair. We compare two (key,value)
   * pairs with the provided key-comparison function D.compare. For example,
   * we may choose to keep a dictionary mapping links to their ranks. In this
   * case, our (key,value) pairs will be (link,rank) pairs, and we compare
   * links using string comparison. *)
  type key = D.key
  type value = D.value
  type pair = key * value

  (* Type definition used for instructor testing. This type is not used for this assignment *)
  type rep =
    | RLeaf
    | RNode of rep * key * value * rep


    (* Type definition for dictionary, which we choose to represent as a AVL Tree.
   * This is almost the same as the binary search tree definition from pset4 and
   * lecture, except we add the invarient that the difference between the height
      of a nodes two subtrees cannot be greater than one.

    Each node contains a left and right subtree, a value and an iteger height *)
  type dict =
    | Leaf
    | Node of dict * int * pair * dict 

  (* TODO How do we represent an empty dictionary with AVL trees? *)
  let empty : dict = raise TODO

  (* TODO Given a dict, return its height. Remeber the height of a
   * node is one greater than the height of its largest subtree. *)
  
  let rec height (d:dict) =
    raise TODO

(* TODO:
   * Implement fold. Read the specification in the DICT signature above. *)
  let rec fold (f: key -> value -> 'a -> 'a) (u: 'a) (d: dict) : 'a =
    raise TODO

(* TODO:
   * Implement the balance method. This method guarentees that every node
   * in the dictionary has a height differenece of one or less between its
   * subtrees. *)
  let rec balanced (d: dict) : bool =
    raise TODO


(* TODO:
   * Implement these to-string functions
   * of_key and of_value are given as anonymous functions to avoid 
   * crashing the program if run while not implemented even if they 
   * are not called (cf of_dict, which is already a function). When 
   * you implement them, you can remove the function wrappers *)
  let string_of_key (k:key) : string = raise TODO
  let string_of_value (v:value) : string = raise TODO
  let rec string_of_dict (d: dict) : string =
      raise TODO

(* Debugging function. This will print out the tree in text format.
   * Use this function to see the actual structure of your 2-3 tree. *
   *
   * e.g.      (4,d)   (6,f)
   *         /       |       \
   *      (2,b)    (4,d)     Leaf
   *      /  \     /   \
   *   Leaf  Leaf Leaf  Leaf
   *
   * string_of_tree will output:
   * Three(Two(Leaf,(2,b),Leaf),(4,d),Two(Leaf,(5,e),Leaf),(6,f),Leaf)
   *
   * Note that this tree is NOT balanced, because all the paths from (6,f)
   * to its leaves do NOT all have the same length. *)
  let rec string_of_tree (d: dict) : string =
        match d with
      | Leaf -> "Leaf"
      | Node (left, height, value, right) -> "Node(" ^ (string_of_tree left) 
        ^ ",(" ^ (string_of_key (fst (value))) ^ "," ^ (string_of_value (snd(value))) ^ "),"
        ^ (string_of_tree right) ^ ")"

  (* Given a dictionary return its height. For a Node with subtrees of
     different heights, the height is one greater than the height of
     the tallest subtree. The height of a Leaf is 0.*)

  (* TODO:
   * Write a lookup function that returns the value of the given key
   * in our dictionary and returns it as an option, or return None
   * if the key is not in our dictionary. *)
  let rec lookup (d: dict) (k: key) : value option =
    raise TODO

  (* TODO:
   * Write a function to test whether a given key is in our dictionary *)
  let member (d: dict) (k: key) : bool =
    raise TODO

  
  (* TODO:
   * Write a function that rotates a tree left. This function
     makes the root of the right subtree the dictionary into the new root of the
     dictionary. The old root of the dictionary becomes a new node as the left
     subtree of the new root, with its original left subtree unchanged
     and the new roots old left subtree as its right subtree*)
  let rotate_left (d: dict) : dict =
    raise TODO

  (* TODO:
   * Write a function that rotates a tree right. This function
     makes the root of the left subtree the dictionary into the new root of the
     dictionary. The old root of the dictionary becomes a new node as the right
     subtree of the new root, with its original right subtree unchanged
     and the new roots old right subtree as its left subtree*)
  let rotate_right (d: dict) : dict =
    raise TODO

(* TODO:
   * Write a function that rebalences a tree after an insert. If difference
   between the subtrees of the dictionary is greater than one, the tree must
   be rebalenced. If the left subtree is too large, a rotate_right must be
   preformed, and a rotate_left if the right_subtree is too large. If the new node
   is inserted into the right subtree of the left subtree or the left subtree of
   the right subtree, then that subtree must be rotated in the opposite direction of
   the tree as a whole. if nessisary, rotate the subtree, then, rotate the entire
   dictionary. *)
  let rebalence (d: dict) (k : key) : dict =
    raise TODO
   
(* Given a AVL tree d, return a new AVL tree which
 * additionally contains the pair (k,v)
 *)
  let rec insert (d: dict) (k: key) (v: value) : dict =
    raise TODO

(* TODO:
   * Write a function that rebalences a tree after a remove. If difference
   between the subtrees of the dictionary is greater than one, the tree must
   be rebalenced. If the left subtree is too small, a rotate_left must be
   preformed, and a rotate_rigth if the right_subtree is too small. If the new node
   was removed from the right subtree of the left subtree or the left subtree of
   the right subtree, then rotate the relevent subtree in the opposite direction.
   of the tree as a whole. Rotate the tree is the approprate direction. *)
  let rec rebalence_remove (d: dict) : dict =
    raise TODO

  (* given a AVL tree d, return an AVL without element k *)
  let rec remove (d: dict) (k: key) : dict =
    raise TODO

(* TODO:
   * Write a function that removes any (key,value) pair from our 
   * dictionary (your choice on which one to remove), and returns
   * as an option this (key,value) pair along with the new dictionary. 
   * If our dictionary is empty, this should return None. *)
  let choose (d: dict) : (key * value * dict) option =
   raise TODO

  (* Generators for randomized testing *)

  (* Generator for keys *)
  let arbitrary_key () : (key QCheck.Gen.t) = D.arbitrary_key ()

  (* Generator for values *)
  let arbitrary_value () : (value QCheck.Gen.t) = D.arbitrary_value ()

  (* Generator for random dictionaries*)
  let arbitrary () : (dict QCheck.Gen.t) =
   (* Create the skeleton for a random tree *)
   let tree_gen = QCheck.Gen.(sized @@ fix
  (fun self (n:int) -> match n with
    | 0 -> return Leaf
    | n ->
      (* Select with equal frequency a tree with either the same height subtrees, *)
      (* a tree with a left subtree one level higher than the right subtree, or a *)
      (* tree with a right subtree on level higher than the left subtree.         *)
      frequency
        [1,  map2 (fun x y -> Node (x, 0, D.gen_pair (), y)) (self (n/2)) (self (n/2));
         1,  map2 (fun x y -> Node (x, 0, D.gen_pair (), y)) (self (n/2)) (self (n/4));
         1, map2 (fun x y -> Node (x, 0, D.gen_pair (), y)) (self (n/4)) (self (n/2))]
    ));
   in
   (* Fill in a tree with random ordered values *)
   let rec fill_in_tree (tree : dict) (min : key) : (key * int * dict) =
      match tree with
      | Leaf -> (min, 1, Leaf)
      | Node (left, h, v, right) ->
        let (min_2, h_2, tree_left) = fill_in_tree left min in
        let value = D.gen_key_gt min_2 () in
        let (min_3, h_3, tree_right) = fill_in_tree right value in
        let h_final = max h_2 h_3 in
        (min_3, h_final + 1, Node (tree_left, h_final, (value, snd v), tree_right))
    in
    QCheck.Gen.(tree_gen >>= (fun tree -> let (_,_,c) = fill_in_tree tree (D.gen_key ()) in return c ))

  (* Return a list of subtrees when given a tree*)
  let shrink (d:dict) : dict list=
      match d with
      | Leaf -> []
      | Node (a, h, v, b) -> a::b::[]

  (* Convert a dict to a testing repersentation *)
  let rec expose (d:dict) : rep =
    match d with
    | Leaf -> RLeaf
    | Node (left, height, (key, value), right) ->
      RNode (expose left, key, value, expose right)

  (* Convert a testing repersentation to a dict*)
  let rec hide_help (r:rep) : (int * dict) =
    match r with
    | RLeaf -> (0, Leaf)
    | RNode (left, key, value, right) ->
      let h1, right_new = hide_help right in
      let h2, left_new = hide_help left in
      ((max h1 h2) + 1, Node (left_new, (max h1 h2) + 1, (key, value), right_new))

  (* Convert a testing repersentation to a dict *)
  let rec hide (r:rep) : dict =
    snd (hide_help r)


  (********************************************************************) 
  (*       TESTS                                                      *)
  (* Section for module tests. We will not be providing tests for this*)
  (* assignment. Instead, you are strongly encouraged to use the       *)
  (* methods provided below to create your own tests. The run_test    *)
  (* takes a property function in the form dict -> bool and the name  *)
  (* of a test, and uses random tests to determine if that property   *)
  (* holds for all cases.                                             *)
  (* EXAMPLE
    run_test (fun d ->
        match choose d with
        | None -> d == empty
        | Some (k,v,d2) ->
          (member d k) && (member d2 k == false) && balanced d2
      ) "Choose";
  *)
  (* This will run random tests to determine if the choose function   *)
  (* will always return a value that was a member of the original dict,*)
  (* and that the dict returned by choose no longer has this value and*)
  (* is balanced. Two additonal functions are provided, run_test_key, *)
  (* which requiers a function of the form dict->key->bool, and       *)
  (* run_test_key_values, which requiers a function of the form       *)
  (* dict->key->value->bool.                                          *)
  (* To run these tests, compile with make test, then run             *)
  (* random_tests.native                                              *)
  (********************************************************************)

   (* Given a property function f : dict->bool, test if that function is true for random dictionaries*)
   let run_test (f : dict -> bool) (s:string) =
     print_string ("Testing "^s^"\n");
     let shrink_fun (x:dict) = QCheck.Iter.of_list (shrink x) in
     let test =
       QCheck.Test.make ~count:1000 ~name:s
       (QCheck.make (arbitrary ()) ~print:string_of_tree ~shrink:shrink_fun) f
     in
     QCheck.Test.check_exn test

   (* Given a property function f : dict->key->bool, test if that function is true
      for random dictionaries and random keys *)
   let run_test_key (f : dict -> key -> bool) (s:string) =
     print_string ("Testing "^s^"\n");
     let pair_test (dict, key) = f dict key in
     let new_gen = QCheck.Gen.map2 (fun dict key -> (dict, key)) (arbitrary ()) (D.arbitrary_key ()) in
     let print_fun (dict, key) = "\n"^(string_of_tree dict)^"\nKey:"^(string_of_key key)^"/n" in
     let shrink_fun (dict, key) =
       QCheck.Iter.of_list
         (List.map (fun sub_dict -> (sub_dict, key)) (shrink dict))
     in
     let test =
       QCheck.Test.make
         ~count:1000
         ~name:s
         (QCheck.make (new_gen) ~print:print_fun ~shrink:shrink_fun)
         pair_test
     in
     QCheck.Test.check_exn test

(* Given a property function f : dict->key->value->bool, test if that function is true
      for random dictionaries, random keys and random values *)
let run_test_key_value (f : dict -> key -> value -> bool) (s:string) =
     print_string ("Testing "^s^"\n");
     let pair_test (dict, key, value) = f dict key value in
     let new_gen =
       QCheck.Gen.map3
         (fun dict key value ->(dict, key, value))
         (arbitrary ())
         (D.arbitrary_key ())
         (D.arbitrary_value ())
     in
     let print_fun (dict, key, value) =
       "/n"^
       (string_of_tree dict)^
       "\nKey:"^
       (string_of_key key)^
       "Value:"^
       (string_of_value value)^
       "/n"
     in
     let shrink_fun (dict, key, value) = QCheck.Iter.of_list (List.map (fun sub_dict -> (sub_dict,key,value)) (shrink dict)) in
     let test =
       QCheck.Test.make
         ~count:1000
         ~name:s
         (QCheck.make (new_gen) ~print:print_fun ~shrink:shrink_fun)
         pair_test
     in
     QCheck.Test.check_exn test
   
(* Run the tests for this module. Please add your tests to this function *)
let run_tests () =
    (* Add tests here *)
    print_string "Tests Passed\n";
    ()


end
*)

(******************************************************************)
(* Make: a functor that creates a DICT by calling our             *)
(* AssocListDict or ALVDict functors                               *)
(******************************************************************)
module Make (D:DICT_ARG) : (DICT with type key = D.key
  with type value = D.value) = 
  (* Change this line to the AVLDict implementation when you are
   * done implementing your AVL trees. *)
  (*AVLDict(D)*)
   AssocListDict(D)
open Format

