(* Definitions for sets. *)

(* An interface for set modules *)
module type SET = 
sig
  type elt  (* type of elements in the set *)
  type set  (* abstract type for the set *)

  val empty : set

  val is_empty : set -> bool

  val insert : elt -> set -> set

  (* same as insert x empty *)
  val singleton : elt -> set

  val union : set -> set -> set
  val intersect : set -> set -> set

  (* remove an element from the set -- if the
   * element isn't present, does nothing. *)
  val remove : elt -> set -> set

  (* returns true iff the element is in the set *)
  val member : set -> elt -> bool

  (* chooses some member from the set, removes it 
   * and returns that element plus the new set.  
   * If the set is empty, returns None. *)
  val choose : set -> (elt * set) option

  (* fold a function across the elements of the set
   * in some unspecified order. *)
  val fold : (elt -> 'a -> 'a) -> 'a -> set -> 'a

  (* functions to convert our types to a string. useful for debugging. *)
  val string_of_set : set -> string
  val string_of_elt : elt -> string

  (* runs our tests. See TESTING EXPLANATION *)
  val run_tests : unit -> unit
end



(* parameter to Set modules -- we must pass in some 
 * type for the elements of a set, a comparison
 * function, and a way to stringify it.
 *)
module type COMPARABLE = 
sig
  type t
  val compare : t -> t -> Order.order
  val string_of_t : t -> string

  (* The functions below are used for testing. See TESTING
   * EXPLANATION *)

  (* Generate a value of type t. The same t is always returned *)
  val gen : unit -> t

  (* Generate a random value of type t. *)
  val gen_random : unit -> t

  (* Generate a t greater than the argument. *)
  val gen_gt : t -> unit -> t

  (* Generate a t less than the argument. *)
  val gen_lt : t -> unit -> t

  (* Generate a t between the two arguments. Return None if no such
   * t exists. *)
  val gen_between : t -> t -> unit -> t option
end



(* An example implementation of our COMPARABLE signature. Use this
 * struct for testing. *)
module IntComparable : COMPARABLE =
struct
  open Order
  type t = int
  let compare x y = if x < y then Less else if x > y then Greater else Eq
  let string_of_t = string_of_int
  let gen () = 0
  let gen_random =
    let _ = Random.self_init () in
    (fun () -> Random.int 10000)
  let gen_gt x () = x + 1
  let gen_lt x () = x - 1
  let gen_between x y () = 
    let (lower, higher) = (min x y, max x y) in
    if higher - lower < 2 then None else Some (higher - 1)
end



(* A simple, list-based implementation of sets. *)
module ListSet(C: COMPARABLE) : (SET with type elt = C.t) = 
struct
  open Order
  type elt = C.t 
  type set = elt list

  (* INVARIANT: sorted, no duplicates *)
  let empty = []
  let is_empty xs = 
    match xs with 
      | [] -> true
      | _ -> false
  let singleton x = [x]
  let rec insert x xs = 
    match xs with 
      | [] -> [x]
      | y::ys -> (match C.compare x y with 
          | Greater -> y::(insert x ys)
          | Eq -> xs
          | Less -> x::xs)

  let union xs ys = List.fold_right insert xs ys
  let rec remove y xs = 
    match xs with 
      | [] -> []
      | x::xs1 -> (match C.compare y x with 
          | Eq -> xs1
          | Less -> xs
          | Greater -> x::(remove y xs1))

  let rec intersect xs ys = 
    match xs, ys with 
      | [], _ -> []
      | _, [] -> []
      | xh::xt, yh::yt -> (match C.compare xh yh with 
          | Eq -> xh::(intersect xt yt)
          | Less -> intersect xt ys
          | Greater -> intersect xs yt)

  let rec member xs x = 
    match xs with 
      | [] -> false
      | y::ys -> (match C.compare x y with
          | Eq -> true
          | Greater -> member ys x
          | Less -> false)

  let choose xs = 
    match xs with 
      | [] -> None
      | x::rest -> Some (x,rest)
  let fold f e = List.fold_left (fun a x -> f x a) e 
    
  let string_of_elt = C.string_of_t
  let string_of_set (s: set) : string = 
    let f = (fun y e -> y ^ "; " ^ C.string_of_t e) in
    "set([" ^ (List.fold_left f "" s) ^ "])"

  let run_tests () = 
    ()

end


(******************************************************************)
(* DictSet: a functor that creates a SET by calling our           *)
(* Dict.Make functor                                              *)
(******************************************************************)
(* TODO Complete this!
module DictSet(C : COMPARABLE) : (SET with type elt = C.t) = 
struct
  module D = Dict.Make(struct

    Add the other requiered fields here!

    let arbitrary_key () = QCheck.Gen.map (fun x -> C.gen_random ()) QCheck.Gen.(nat)
    let arbitrary_value () = QCheck.Gen.return ()
end)
 
  type elt = D.key
  type set = D.dict
  let empty = ??

   Implement the rest of the set functions!

  let string_of_elt = D.string_of_key
  let string_of_set s = D.string_of_dict s
  
  (********************************************************************) 
  (*       TESTS                                                      *)
  (* Section for module tests. We will not be providing tests for this*)
  (* assignment. Instead, you are strongly encouraged to use the       *)
  (* methods provided below to create your own tests. The run_test    *)
  (* takes a property function in the form set -> bool and the name  *)
  (* of a test, and uses random tests to determine if that property   *)
  (* holds for all cases.                                             *)
  (* EXAMPLE
    run_test (fun set ->
        match choose set with
        | None -> is_empty set
        | Some (v,set2) ->
          (member set v) && (member set2 v == false)
      ) "Choose";
  *)
  (* This will run random tests to determine if the choose function   *)
  (* will always return a value that was a member of the original set, *)
  (* and that the set returned by choose no longer has this value.    *)
  (* Two additonal functions are provided, run_test_elt,              *)
  (* which requiers a function of the form set->elt->bool, and        *)
  (* run_test_set_elt, which requiers a function of the form          *)
  (* dict->dict->value->bool.                                         *)
  (* To run these tests, compile with make test, then run             *)
  (* random_tests.native                                              *)
  (********************************************************************)


 (* Given a property function f : set->bool, test if that function is true for random sets*)
  let run_test (f : set -> bool) (s:string) =
     print_string ("Testing "^s^"\n");
     let shrink_fun (x:set) = QCheck.Iter.of_list (D.shrink x) in
     let test =
       QCheck.Test.make ~count:1000 ~name:s
       (QCheck.make (D.arbitrary ()) ~print:D.string_of_dict ~shrink:shrink_fun) f
     in
     (QCheck.Test.check_exn test)
 
  (* Given a property function f : set->elt->bool, test if that function is true
      for random sets and random keys *)
   let run_test_elt (f : set -> elt -> bool) (s:string) = 
     print_string ("Testing "^s^"\n");
     let pair_test (set, key) = f set key in
     let new_gen = QCheck.Gen.map2 (fun set key ->
         (set, key)) (D.arbitrary ()) (D.arbitrary_key ())
     in
     let print_fun (set, key) = "\n"^(D.string_of_dict set)^"\nKey:"^(string_of_elt key)^"/n" in
     let shrink_fun (set, key) =
       QCheck.Iter.of_list
         (List.map (fun sub_set -> (sub_set, key)) (D.shrink set))
     in
     let test =
       QCheck.Test.make
         ~count:1000
         ~name:s
         (QCheck.make (new_gen) ~print:print_fun ~shrink:shrink_fun)
         pair_test
     in
     (QCheck.Test.check_exn test)

(* Given a property function f : set->set->elt->bool, test if that function is true
      for random sets and random keys *)
let run_test_set_elt (f : set -> set -> elt -> bool) (s:string) =
     print_string ("Testing "^s^"\n");
     let pair_test (set1, set2, key) = f set1 set2 key in
     let new_gen =
       QCheck.Gen.map3
         (fun set1 set2 key ->(set1, set2, key))
         (D.arbitrary ())
         (D.arbitrary ())
         (D.arbitrary_key ())
     in
     let print_fun (set1, set2, key) =
       "\n"^
       (string_of_set set1)^
       "\n"^
       (string_of_set set2)^
       "\nKey:"^
       (string_of_elt key)^
       "\n"
     in
     let shrink_fun (set1,set2, key) =
       QCheck.Iter.of_list (
         let s1 = D.shrink set1 in
         let s2 = D.shrink set2 in
         let rec shr_help sub1 sub2 =
           match (sub1, sub2) with
           | ([], _) -> []
           | (_, []) -> [] 
           | (a::b, c::d) -> (a,c,key)::(shr_help b d)
         in
         shr_help s1 s2)
     in 
     let test =
       QCheck.Test.make
         ~count:1000
         ~name:s
         (QCheck.make (new_gen) ~print:print_fun ~shrink:shrink_fun)
         pair_test
     in
     QCheck.Test.check_exn test

  (* add your test functions to run_tests *)
  let run_tests () = 
    ()
end

*)
(******************************************************************)
(* Run our tests.                                                 *)
(******************************************************************)

(* Create a set of ints using our ListSet functor. *)
module IntListSet = ListSet(IntComparable) ;;
IntListSet.run_tests();;

(* Create a set of ints using our DictSet functor
 * 
 * Uncomment out the lines below when you are ready to test your
 * 2-3 dict set implementation *)
(*
module IntDictSet = DictSet(IntComparable) ;;
IntDictSet.run_tests();;
*)
(******************************************************************)
(* Make: a functor that creates a SET by calling our              *)
(* ListSet or DictSet functors                                    *)
(******************************************************************)
module Make(C : COMPARABLE) : (SET with type elt = C.t) = 
  (* Change this line to use our dictionary implementation when your are 
   * finished. *)
  ListSet (C)
  (* DictSet (C) *)

