open Util ;;    
open CrawlerServices ;;
open Order ;;
open Nodescore ;;
open Graph ;;


(* Dictionaries mapping links to their ranks. Higher is better. *)
module RankDict = Dict.Make(
  struct 
    type key = link
    type value = float
    let compare = link_compare
    let string_of_key = string_of_link
    let string_of_value = string_of_float
    let gen_key () = {host=""; port=0; path=""}
    let gen_key_gt x () = gen_key ()
    let gen_key_lt x () = gen_key ()
    let gen_key_random () = gen_key ()
    let gen_key_between x y () = None
    let gen_value () = 0.0
    let gen_pair () = (gen_key(),gen_value())
    let arbitrary_key () = QCheck.Gen.map (fun x -> gen_key ()) QCheck.Gen.(nat)
    let arbitrary_value () = QCheck.Gen.map (fun (x : value) -> x) QCheck.Gen.(float)
        
  end)

module PageSet = Myset.Make(
  struct 
    type t = page
    let compare = (fun a b -> link_compare (a.url) (b.url))
    let string_of_t = string_of_page
    let gen () = {url={host=""; port=0; path=""}; links=[]; words=[]}
    let gen_lt x () = gen ()
    let gen_gt y () = gen ()
    let gen_random () = gen ()
    let gen_between x y () = None
  end)

module LinkSet = Myset.Make(
  struct 
    type t = link
    let compare = link_compare
    let string_of_t = string_of_link
    let gen () = {host=""; port=0; path=""}
    let gen_lt x () = gen ()
    let gen_gt y () = gen ()
    let gen_random () = gen ()
    let gen_between x y () = None
  end)

module PageGraph = Graph (
  struct
    type node = link
    let compare = link_compare
    let string_of_node = string_of_link
    let gen () = {host=""; port=0; path=""}
  end)

module PageScore = NodeScore (
  struct
    type node = link
    let string_of_node = string_of_link
    let compare = link_compare
    let gen () = {host=""; port=0; path=""}
  end)

(* Given a bunch of pages, convert them to a graph *)
let graph_of_pages (pages : PageSet.set) : PageGraph.graph =
  (* Only want graph nodes for pages we actually crawled *)
  let crawled_links = 
    PageSet.fold (fun page s -> LinkSet.insert page.url s)
      LinkSet.empty pages 
  in
  let add_links page graph =
    let add_link g dst =
      if LinkSet.member crawled_links dst then
        PageGraph.add_edge g page.url dst
      else g 
    in
      List.fold_left add_link graph page.links
  in
    PageSet.fold add_links PageGraph.empty pages
      
(* The rest of the world wants a RankDict, not a NodeScore. *)

let dict_of_ns (ns : PageScore.node_score_map) : RankDict.dict =
  PageScore.fold (fun node score r -> RankDict.insert r node score) 
    RankDict.empty ns

(* A type for modules that can compute nodescores from graphs *)
module type RANKER = 
sig
  module G: GRAPH
  module NS: NODE_SCORE
  val rank : G.graph -> NS.node_score_map
end


(* Each node's rank is equal to the number of pages that link to it. *)
module InDegreeRanker  (GA: GRAPH) (NSA: NODE_SCORE with module N = GA.N) : 
  (RANKER with module G = GA with module NS = NSA) =
struct
  module G = GA
  module NS = NSA
  let rank (g : G.graph) = 
    let add_node_edges ns node =
      let neighbors = match G.neighbors g node with
        | None -> []
        | Some xs -> xs
      in
        List.fold_left (fun ns' neighbor -> NS.add_score ns' neighbor 1.0) 
          ns neighbors 
    in
    let nodes = (G.nodes g) in
      List.fold_left add_node_edges (NS.zero_node_score_map nodes) nodes
end



(*****************************************************************)
(* Random Walk Ranker                                            *)
(*****************************************************************)

module type WALK_PARAMS =
sig
  (* Should we randomly jump somewhere else occasionally? 
    if no, this should be None.  Else it should be the probability of 
    jumping on each step *)
  val do_random_jumps : float option
end


(*****************************************************************)
(* Eigenvalue Ranker                                             *)
(*****************************************************************)

module EigenvalueRanker (GA: GRAPH) (NSA: NODE_SCORE with module N = GA.N) 
  (P : WALK_PARAMS) : 
  (RANKER with module G = GA with module NS = NSA) =
struct
  module G = GA
  module NS = NSA

  let dot_product v1 v2 =
   List.fold_left2 (fun sum val1 val2 -> sum +. (val1 *. val2)) 0.0 v1 v2

  let mult m1 v1 =
   List.fold_right (fun list new_list -> (dot_product list v1)::new_list) m1 []

  let scale d m =
   List.map (fun v -> v *. d) m

  let add_const d m =
   List.map (fun v -> v +. d) m

  let evd v1 v2 = 
   sqrt (List.fold_left2 (fun sum val1 val2 -> sum +. ((val1 -. val2) *.(val1 -. val2))) 0.0 v1 v2)

  let check_link l g =
   let nodes = (G.nodes g) in
   let n = List.length l in
   let v = List.init n (fun x -> 1.0) in
   let c = (dot_product (mult l v) v)-.(float_of_int n) in
   if c > 0.001 || c < -0.001 then failwith "There is an error in the link matrix"
   else
   let rec rand_check rep =
      if rep<0 then () else
      let i = Random.int n in
      let j = Random.int n in
      let v1 = List.init n (fun x -> if x = (i) then 1.0 else 0.0) in
      let v2 = List.init n (fun x -> if x = (j) then 1.0 else 0.0) in
      let index = (dot_product (mult l v1) v2) in
      if index > 0.0 then
        let neighbors = match (G.neighbors g (List.nth nodes i)) with None -> [] | Some x -> (match x with
            | [] -> nodes
            | _ -> x) in
        if (List.mem (List.nth nodes j) neighbors) then rand_check (rep - 1)
        else failwith "There is an error in the link matrix"
   
      else
        let neighbors = match (G.neighbors g (List.nth nodes i)) with None -> [] | Some x -> (match x with
            | [] -> nodes
            | _ -> x) in
        if ((List.mem (List.nth nodes j) neighbors) = false) then rand_check (rep - 1)
        else failwith "There is an error in the link matrix"
   in
   rand_check 100

  (*matrix computation of pagerank*)
  let rank (g : G.graph) =
    let nodes = (G.nodes g) in
    let size = List.length nodes in
    let jump = match  P.do_random_jumps with None -> 0.0 | Some x -> x in
    (* Link matirix, cell is equal to 1 divided by the number of links in that row if there is a
     * link between the row and the column, and 0 otherwise. Rows with no links are changed to like
     * with every other row.*)
    let links = List.fold_right (fun node list->
      (List.fold_right (fun x l->
       let n = match (G.neighbors g x) with
         | None -> nodes
         | Some x -> (match x with
            | [] -> nodes
            | _ -> x) in
     (if (List.mem node n)
       then 1.0/.(float_of_int (List.length n))
       else 0.0
     )::l) nodes [])::list) nodes [] in
   let _ = check_link links g in
   let r0 = List.init size (fun i -> 1.0/.(float_of_int size)) in
   let jump_value = jump/. (float_of_int (size)) in
   (*Loop body. Multiplys the link matrix by the rank vector, and adds the constant for the
    * jump in the process.*)
   let rec recurse n r =
     let r2 = r in
     if n<=0 then r else
     let r' = add_const (jump_value) (scale (1.0 -. jump) (mult links r2)) in
     if evd r' r < (sqrt (dot_product r r))/.1000.0 then
   r'
     else
     recurse (n-1) (r')
   in
   let rfinal = recurse 1000 r0 in
   List.fold_left2 (fun l n x -> NS.add_score l n x) (NS.zero_node_score_map nodes) nodes rfinal

end


(*******************  TESTS BELOW  *******************)

module TestInDegreeRanker =
struct 
  module G = NamedGraph
  let g = G.add_edge G.empty "a" "b";;
  let g2 = G.add_edge g "a" "c";;
  
  module NS = NodeScore (struct
                           type node = string 
                           let compare = string_compare
                           let string_of_node = fun x -> x
                           let gen () = ""
                         end);;

  module Ranker = InDegreeRanker (G) (NS);;
  let ns = Ranker.rank g2;;
  (* let _ = Printf.printf "NS: %s\n" (NS.string_of_node_score_map ns) ;; *)
  assert ((NS.get_score ns "a") = Some 0.0);;
  assert ((NS.get_score ns "b") = Some 1.0);;
  assert ((NS.get_score ns "c") = Some 1.0);;
  assert ((NS.get_score ns "d") = None);;

  let g3 = G.add_edge g2 "b" "c";;
  let ns2 = Ranker.rank g3;;
  assert ((NS.get_score ns2 "a") = Some 0.0);;
  assert ((NS.get_score ns2 "b") = Some 1.0);;
  assert ((NS.get_score ns2 "c") = Some 2.0);;

end


module TestEigenvalueRanker =
struct 
  module G = NamedGraph
  let g = G.from_edges [("a","b") ; 
                        ("b","c") ;
                        ("c","d") ;
                        ("d","e") ;
                        ("e","f") ;
                        ("a","g") ;
                        ("g","a")]
  
  module NS = NodeScore (struct
                           type node = string 
                           let compare = string_compare
                           let string_of_node = fun x -> x
                           let gen () = ""
                         end);;

  module Ranker = EigenvalueRanker (G) (NS) 
    (struct 
       let do_random_jumps = None
     end)

  let ns = Ranker.rank g
  let _ = Printf.printf "Testing RandomWalkRanker:\n NS: %s\n" 
    (NS.string_of_node_score_map ns) 

end



